/*
 * Copyright 2005 James Cleverdon, IBM.
 * Subject to the GNU Public License, v.2
 *
 * gsi2irq.c:
 *
 * IRQ and vector compression/sharing routines for i386 and x86-64 by
 * James Cleverdon from a patch by Natalie Protasevich and
 * architecture code from io_apic.c
 */
#include <linux/config.h>
#include <asm/smp.h>
#include <linux/irq.h>
#include <asm/hw_irq.h>

#ifdef	CONFIG_X86_64

#include <asm/irq.h>

#else	/* X86 */

#include <irq_vectors.h>

#endif	/* X86 */


#ifdef CONFIG_X86_IO_APIC

extern int __assign_irq_vector(int irq);

static int next_irq = 16;

static u8 gsi_2_irq[NR_IRQ_VECTORS] = { [0 ... NR_IRQ_VECTORS-1] = 0xFF };

static spinlock_t gsi_irq_lock = SPIN_LOCK_UNLOCKED;

/* #define DEBUG_GSI_IRQ_SHARING	1 */

#ifdef DEBUG_GSI_IRQ_SHARING

#undef KERN_INFO	/* Raise printk level. */
#define KERN_INFO	KERN_ERR

#define DEBUG_GSI_FORCE_SHARE_N	3
static int gsi_irq_dbg_cnt = DEBUG_GSI_FORCE_SHARE_N;
static int gsi_irq_dbg_irq = 16;

#endif	/* DEBUG_GSI_IRQ_SHARING */


/*
 * gsi_irq_sharing -- Name overload!  "irq" can be either a legacy IRQ
 * in the range 0-15, a linux IRQ in the range 0-223, or a GSI number
 * from ACPI, which can reach 800 in large boxen.
 *
 * Compact the sparse GSI space into a sequential IRQ series and reuse
 * vectors if possible.
 */
int gsi_irq_sharing(int gsi)
{
	int i, tries, vector;
	unsigned long flags;

	BUG_ON(gsi >= NR_IRQ_VECTORS);

	if (platform_legacy_irq(gsi)) {
		gsi_2_irq[gsi] = gsi;
		return gsi;
	}

	if (gsi_2_irq[gsi] != 0xFF)
		return (int)gsi_2_irq[gsi];

	tries = NR_IRQS;
  try_again:
	vector = __assign_irq_vector(gsi);

	/*
	 * Sharing vectors means sharing IRQs, so scan irq_vectors for previous
	 * use of vector and if found, return that IRQ.  However, we never want
	 * to share legacy IRQs, which usually have a different trigger mode
	 * than PCI.
	 */
	for (i = 0; i < NR_IRQS; i++)
		if (IO_APIC_VECTOR(i) == vector)
			break;
	if (platform_legacy_irq(i)) {
		if (--tries >= 0)
			goto try_again;
		panic("gsi_irq_sharing: didn't find an IRQ using vector 0x%02X for GSI %d", vector, gsi);
	}
	if (i < NR_IRQS) {
		gsi_2_irq[gsi] = i;
		printk(KERN_INFO "GSI %d sharing vector 0x%02X and IRQ %d\n",
				gsi, vector, i);
		return i;
	}

	spin_lock_irqsave(&gsi_irq_lock, flags);
#ifdef DEBUG_GSI_IRQ_SHARING
	if (--gsi_irq_dbg_cnt < 0) {
		/* Debug:  Force sharing on 1 of N IRQs.  N must be > 1. */
		gsi_irq_dbg_cnt = DEBUG_GSI_FORCE_SHARE_N;
		i = gsi_irq_dbg_irq++;
		if (gsi_irq_dbg_irq >= next_irq)
			gsi_irq_dbg_irq = 16;
		gsi_2_irq[gsi] = i;
		spin_unlock_irqrestore(&gsi_irq_lock, flags);
		printk(KERN_INFO "GSI %d debug-share vector 0x%02X and IRQ %d\n",
				gsi, IO_APIC_VECTOR(i), i);
		return i;
	}
#endif	/* DEBUG_GSI_IRQ_SHARING */
	i = next_irq++;
	BUG_ON(i >= NR_IRQS);
	gsi_2_irq[gsi] = i;
	IO_APIC_VECTOR(i) = vector;
	vector_irq[vector] = i;
	spin_unlock_irqrestore(&gsi_irq_lock, flags);
	printk(KERN_INFO "GSI %d assigned vector 0x%02X and IRQ %d\n",
			gsi, vector, i);
	return i;
}

#else	/* CONFIG_X86_IO_APIC */

/* No compression needed if no I/O APICs. */

int gsi_irq_sharing(int gsi)
{
	return gsi;
}

#endif	/* CONFIG_X86_IO_APIC */
