#include <linux/init.h>
#include <linux/sched.h>

static struct sched_group sched_group_cpus[NR_CPUS];
static struct sched_group sched_group_phys[NR_CPUS];
static DEFINE_PER_CPU(struct sched_domain, cpu_domains);
static DEFINE_PER_CPU(struct sched_domain, phys_domains);
#ifdef CONFIG_NUMA_SCHED
static struct sched_group sched_group_nodes[MAX_NUMNODES];
static DEFINE_PER_CPU(struct sched_domain, node_domains);
void __init arch_init_sched_domains(void)
{
	int i;
	struct sched_group *first_node = NULL, *last_node = NULL;
	struct sched_group *first = NULL, *last = NULL;

	/* Set up domains */
	for_each_cpu(i) {
		int node = cpu_to_node(i);
		cpumask_t nodemask = node_to_cpumask(node);
		struct sched_domain *node_sd = &per_cpu(node_domains, i);
		struct sched_domain *cpu_sd = &per_cpu(cpu_domains, i);
		struct sched_domain *phys_sd = &per_cpu(phys_domains, i);

#ifdef CONFIG_SCHED_SMT
		*cpu_sd = SD_SIBLING_INIT;
		cpu_sd->span = cpu_sibling_map[i];
		cpu_sd->groups = &sched_group_cpus[i];
		cpu_sd->parent = phys_sd;
#endif

		*phys_sd = SD_CPU_INIT;
		phys_sd->span = nodemask;
		phys_sd->groups = &sched_group_phys[first_cpu(cpu_sd->span)];
		phys_sd->parent = node_sd;


		*node_sd = SD_NODE_INIT;
#ifdef CONFIG_IA64
		/* increase min load balancing interval for large systems */
		if (num_online_cpus() > 64) {
			node_sd->min_interval = 8 * num_online_cpus() / 32;
			node_sd->balance_interval = 1 * \
						(63 + num_online_cpus()) / 64;
		}
#endif
		node_sd->span = cpu_online_map;
		node_sd->groups = &sched_group_nodes[cpu_to_node(i)];

	}

#ifdef CONFIG_SCHED_SMT
	/* Set up CPU (siblings) groups */
	for_each_cpu(i) {
		struct sched_domain *cpu_domain = &per_cpu(cpu_domains, i);
		int j;
		first = last = NULL;

		if (i != first_cpu(cpu_domain->span))
			continue;

		for_each_cpu_mask(j, cpu_domain->span) {
			struct sched_group *cpu = &sched_group_cpus[j];
			
			cpus_clear(cpu->cpumask);
			cpu_set(j, cpu->cpumask);
			cpu->cpu_power = SCHED_LOAD_SCALE;

			if (!first)
				first = cpu;
			if (last)
				last->next = cpu;
			last = cpu;
		}
		last->next = first;
	}
#endif

	/* Set up groups */
	for (i = 0; i < MAX_NUMNODES; i++) {
		cpumask_t tmp = node_to_cpumask(i);
		cpumask_t nodemask;
		struct sched_group *node = &sched_group_nodes[i];
		int j;

		cpus_and(nodemask, tmp, cpu_online_map);

		if (cpus_empty(nodemask))
			continue;

		node->cpumask = nodemask;

		first = last = NULL;

		for_each_cpu_mask(j, node->cpumask) {
			struct sched_group *cpu = &sched_group_phys[j];
#ifdef CONFIG_SCHED_SMT
			struct sched_domain *cpu_domain = &per_cpu(cpu_domains, j);
			
			if (j != first_cpu(cpu_domain->span))
				continue;

			cpu->cpumask = cpu_domain->span;
#else
			cpus_clear(cpu->cpumask);
			cpu_set(j, cpu->cpumask);
#endif
			cpu->cpu_power = SCHED_LOAD_SCALE + SCHED_LOAD_SCALE*(cpus_weight(cpu->cpumask)-1) / 10;
			node->cpu_power += cpu->cpu_power;

			if (!first)
				first = cpu;
			if (last)
				last->next = cpu;
			last = cpu;
		}
		last->next = first;

		if (!first_node)
			first_node = node;
		if (last_node)
			last_node->next = node;
		last_node = node;
	}
	last_node->next = first_node;

	mb();
	for_each_cpu(i) {
#ifdef CONFIG_SCHED_SMT
		struct sched_domain *cpu_sd = &per_cpu(cpu_domains, i);
#else
		struct sched_domain *cpu_sd = &per_cpu(phys_domains, i);
#endif
		cpu_attach_domain(cpu_sd, i);
	}
}

#else /* !CONFIG_NUMA_SCHED */
void __init arch_init_sched_domains(void)
{
	int i;
	struct sched_group *first = NULL, *last = NULL;

	/* Set up domains */
	for_each_cpu(i) {
		struct sched_domain *phys_sd = &per_cpu(phys_domains, i);
#ifdef CONFIG_SCHED_SMT
		struct sched_domain *cpu_sd = &per_cpu(cpu_domains, i);

		*cpu_sd = SD_SIBLING_INIT;
		cpu_sd->span = cpu_sibling_map[i];
		cpu_sd->parent = phys_sd;
		cpu_sd->groups = &sched_group_cpus[i];
#endif
		*phys_sd = SD_CPU_INIT;
		phys_sd->span = cpu_online_map;
#ifdef CONFIG_SCHED_SMT
		phys_sd->groups = &sched_group_phys[first_cpu(cpu_sd->span)];
#else
		phys_sd->groups = &sched_group_phys[i];
#endif
	}

#ifdef CONFIG_SCHED_SMT
	/* Set up CPU (siblings) groups */
	for_each_cpu(i) {
		struct sched_domain *cpu_domain = &per_cpu(cpu_domains, i);
		int j;
		first = last = NULL;

		if (i != first_cpu(cpu_domain->span))
			continue;

		for_each_cpu_mask(j, cpu_domain->span) {
			struct sched_group *cpu = &sched_group_cpus[j];
			
			cpus_clear(cpu->cpumask);
			cpu_set(j, cpu->cpumask);
			cpu->cpu_power = SCHED_LOAD_SCALE;

			if (!first)
				first = cpu;
			if (last)
				last->next = cpu;
			last = cpu;
		}
		last->next = first;
	}

	first = last = NULL;
#endif
	for_each_cpu(i) {
		struct sched_group *cpu = &sched_group_phys[i];
#ifdef CONFIG_SCHED_SMT
		struct sched_domain *cpu_domain = &per_cpu(cpu_domains, i);

		if (i != first_cpu(cpu_domain->span))
			continue;

		cpu->cpumask = cpu_domain->span;
#else
		cpus_clear(cpu->cpumask);
		cpu_set(i, cpu->cpumask);
#endif
		/*
		 * Make each extra sibling increase power by 10% of
		 * the basic CPU. This is very arbitrary.
		 */
		cpu->cpu_power = SCHED_LOAD_SCALE + SCHED_LOAD_SCALE*(cpus_weight(cpu->cpumask)-1) / 10;

		if (!first)
			first = cpu;
		if (last)
			last->next = cpu;
		last = cpu;
	}
	last->next = first;

	mb(); /* domains were modified outside the lock */
	for_each_cpu(i) {
#ifdef CONFIG_SCHED_SMT
		struct sched_domain *cpu_sd = &per_cpu(cpu_domains, i);
#else
		struct sched_domain *cpu_sd = &per_cpu(phys_domains, i);
#endif
		cpu_attach_domain(cpu_sd, i);
	}
}
#endif
