/*
 * ipmi_kcompat.h
 *
 * Compatability functions for new 2.6 kernel functions on older 2.6 kernels
 *
 * Copyright (c) 2005 Dell, Inc.
 *   by Matt Domsch <Matt_Domsch@dell.com>
 *
 * This file is released under the GPLv2
 */

#ifndef __LINUX_IPMI_KCOMPAT_H
#define __LINUX_IPMI_KCOMPAT_H

#include <linux/config.h>
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/notifier.h>
#include <linux/sched.h>
#include <asm/hardirq.h>


#ifndef __iomem
#define __iomem
#endif

#ifndef __user
#define __user
#endif

#ifndef DEFINE_SPINLOCK
#define DEFINE_SPINLOCK(x) spinlock_t x = SPIN_LOCK_UNLOCKED
#endif

/**
 * pci_get_class - begin or continue searching for a PCI device by class
 * @class: search for a PCI device with this class designation
 * @from: Previous PCI device found in search, or %NULL for new search.
 *
 * Iterates through the list of known PCI devices.  If a PCI device is
 * found with a matching @class, the reference count to the device is
 * incremented and a pointer to its device structure is returned.
 * Otherwise, %NULL is returned.
 * A new search is initiated by passing %NULL to the @from argument.
 * Otherwise if @from is not %NULL, searches continue from next device
 * on the global list.  The reference count for @from is always decremented
 * if it is not %NULL.
 *
 * Note: this implementation differs from the standard pci_get_class()
 * routine because pci_find_class() drops pci_bus_lock before returning,
 * so we do the pci_dev_put(from) and pci_dev_get(dev) without holding
 * pci_bus_lock.  Is this a problem?  I hope not.
 */
static inline struct pci_dev *pci_get_class(unsigned int class, struct pci_dev *from)
{
        struct pci_dev *dev;

        WARN_ON(in_interrupt());
	dev = pci_find_class(class, from);
	pci_dev_put(from);
	dev = pci_dev_get(dev);
        return dev;
}

/* Internal kernel extensions */
#ifndef FMODE_LSEEK
#define FMODE_LSEEK	4
#endif
#ifndef FMODE_PREAD
#define FMODE_PREAD	8
#endif
#ifndef FMODE_PWRITE
#define FMODE_PWRITE	FMODE_PREAD	/* These go hand in hand */
#endif

/*
 * This is used by subsystems that don't want seekable
 * file descriptors
 */
static inline int nonseekable_open(struct inode *inode, struct file *filp)
{
	filp->f_mode &= ~(FMODE_LSEEK | FMODE_PREAD | FMODE_PWRITE);
	return 0;
}

/*
 * A pointer passed in from user mode. This should not
 * be used for syscall parameters, just declare them
 * as pointers because the syscall entry code will have
 * appropriately comverted them already.
 */
typedef u32             compat_uptr_t;

static inline void __user *compat_ptr(compat_uptr_t uptr)
{
        return (void __user *)(unsigned long)uptr;
}

#if defined(CONFIG_X86_64)
static __inline__ void __user *compat_alloc_user_space(long len)
{
        struct pt_regs *regs = (void *)current->thread.rsp0 - sizeof(struct pt_regs);
        return (void __user *)regs->rsp - len;
}
#endif

#define synchronize_sched synchronize_kernel

#ifdef CONFIG_WATCHDOG_NOWAYOUT
#define WATCHDOG_NOWAYOUT	1
#else
#define WATCHDOG_NOWAYOUT	0
#endif

/* CTL_DEV names: */
enum {
        DEV_IPMI=7,
};

/* /proc/sys/dev/ipmi */
enum {
        DEV_IPMI_POWEROFF_POWERCYCLE=1,
};

/*
 * Clean way to return from the notifier and stop further calls.
 */
#ifndef NOTIFY_STOP
#define NOTIFY_STOP		(NOTIFY_OK|NOTIFY_STOP_MASK)
#endif

static inline signed long schedule_timeout_interruptible(signed long timeout)
{
       __set_current_state(TASK_INTERRUPTIBLE);
       return schedule_timeout(timeout);
}

#endif /* __LINUX_IPMI_KCOMPAT_H */
