/*
 *
 * linux/drivers/s390/scsi/zfcp_stat.c
 *
 * FCP adapter driver for IBM eServer zSeries
 *
 * Statistics
 *
 * (C) Copyright IBM Corp. 2005
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define ZFCP_STAT_REVISION "$Revision: 1.12.2.1 $"

#include <linux/statistic.h>
#include <linux/ctype.h>
#include "zfcp_ext.h"

#define ZFCP_LOG_AREA			ZFCP_LOG_AREA_OTHER

static void zfcp_statistic_pull(struct statistic_interface *interface)
{
	struct zfcp_adapter *a = interface->private;
	struct fsf_qtcb_bottom_config *cd;
	struct fsf_qtcb_bottom_port *pd;
	int ret;

	if (!a->adapter_features & FSF_FEATURE_MEASUREMENT_DATA)
		return;

	cd = kmalloc(sizeof(*cd), GFP_KERNEL);
	pd = kmalloc(sizeof(*pd), GFP_KERNEL);
	if (!cd || !pd)
		return;
	memset(cd, 0x0, sizeof *cd);
	memset(pd, 0x0, sizeof *pd);

	ret = zfcp_fsf_exchange_config_data_sync(a, cd);
	if (!ret) {
		statistic_set(a->stat_vhba_input, 1, cd->input_requests);
		statistic_set(a->stat_vhba_output, 1, cd->output_requests);
		statistic_set(a->stat_vhba_control, 1, cd->control_requests);
		statistic_set(a->stat_vhba_received, 1, cd->input_mb);
		statistic_set(a->stat_vhba_sent, 1, cd->output_mb);
		statistic_set(a->stat_vhba_active, 1, cd->active_time);
	}

	ret = zfcp_fsf_exchange_port_data(NULL, a, pd);
	if (!ret) {
		statistic_set(a->stat_phba_processor, 1, pd->cpu_utilization);
		statistic_set(a->stat_phba_bus, 1, pd->bus_utilization);
		statistic_set(a->stat_phba_adapter, 1, pd->adapter_utilization);
	}

	kfree(cd);
	kfree(pd);
}

int zfcp_adapter_statistic_register(struct zfcp_adapter *adapter)
{
	int retval = 0;
	char name[14];

	sprintf(name, "zfcp-%s", zfcp_get_busid_by_adapter(adapter));
	if (statistic_interface_create(&adapter->stat_if, name) )
		return -EINVAL;

	retval = statistic_create(&adapter->stat_qdio_outb_full, 
					adapter->stat_if,
					"occurrence_qdio_outb_full",
					"sbals_left/incidents");
	if (retval)
		goto err_out;	
	
	retval = statistic_create(&adapter->stat_qdio_outb, 
					adapter->stat_if,
					"util_qdio_outb",
					"slots-occupied/incidents");
	if (retval)
		goto err_out;
	
	retval = statistic_create(&adapter->stat_qdio_inb, 
					adapter->stat_if,
					"util_qdio_inb", 
					"slots-occupied/incidents");
	if (retval)
		goto err_out;
	
	retval = statistic_create(&adapter->stat_low_mem_scsi, 
					adapter->stat_if,
					"occurrence_low_mem_scsi", 
					"-/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&adapter->stat_erp, 
					adapter->stat_if,
					"occurrence_erp", "results/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&adapter->stat_vhba_input,
					adapter->stat_if,
					"virt_hba_input", "-/requests");
	if (retval)
		goto err_out;

	retval = statistic_create(&adapter->stat_vhba_output,
					adapter->stat_if,
					"virt_hba_output", "-/requests");
	if (retval)
		goto err_out;

	retval = statistic_create(&adapter->stat_vhba_control,
					adapter->stat_if,
					"virt_hba_control", "-/requests");
	if (retval)
		goto err_out;

	retval = statistic_create(&adapter->stat_vhba_sent,
					adapter->stat_if,
					"virt_hba_sent", "-/megabytes");
	if (retval)
		goto err_out;

	retval = statistic_create(&adapter->stat_vhba_received,
					adapter->stat_if,
					"virt_hba_received", "-/megabytes");
	if (retval)
		goto err_out;

	retval = statistic_create(&adapter->stat_vhba_active,
					adapter->stat_if,
					"virt_hba_active", "-/seconds");
	if (retval)
		goto err_out;

	retval = statistic_create(&adapter->stat_phba_processor,
					adapter->stat_if,
					"phys_hba_processor", "-/percent");
	if (retval)
		goto err_out;

	retval = statistic_create(&adapter->stat_phba_bus,
					adapter->stat_if,
					"phys_hba_bus", "-/percent");
	if (retval)
		goto err_out;

	retval = statistic_create(&adapter->stat_phba_adapter,
					adapter->stat_if,
					"phys_hba_adapter", "-/percent");
	if (retval)
		goto err_out;
	
	statistic_define_value(adapter->stat_qdio_outb_full,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);
	
	statistic_define_range(adapter->stat_qdio_outb,
				0, QDIO_MAX_BUFFERS_PER_Q - 1);
	
	statistic_define_range(adapter->stat_qdio_inb,
				0, QDIO_MAX_BUFFERS_PER_Q - 1);
	
	statistic_define_value(adapter->stat_low_mem_scsi, STATISTIC_RANGE_MIN,
				STATISTIC_RANGE_MAX, STATISTIC_DEF_MODE_INC);

	statistic_define_value(adapter->stat_erp,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	statistic_define_value(adapter->stat_vhba_input,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	statistic_define_value(adapter->stat_vhba_output,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	statistic_define_value(adapter->stat_vhba_received,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	statistic_define_value(adapter->stat_vhba_sent,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	statistic_define_value(adapter->stat_vhba_active,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	statistic_define_value(adapter->stat_phba_processor,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	statistic_define_value(adapter->stat_phba_bus,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	statistic_define_value(adapter->stat_phba_adapter,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	adapter->stat_if->pull = zfcp_statistic_pull;
	adapter->stat_if->private = adapter;

	return 0;
err_out:
	statistic_interface_remove(&adapter->stat_if);
	return retval;
}

int zfcp_adapter_statistic_unregister(struct zfcp_adapter *adapter)
{
	return statistic_interface_remove(&adapter->stat_if);
}

int zfcp_unit_statistic_register(struct zfcp_unit *unit)
{
	int retval = 0;
	char name[64];

	atomic_set(&unit->read_num, 0);
	atomic_set(&unit->write_num, 0);

	sprintf(name, "zfcp-%s-0x%016Lx-0x%016Lx",
		zfcp_get_busid_by_unit(unit), unit->port->wwpn, unit->fcp_lun);

	if (statistic_interface_create(&unit->stat_if, name))
		return -EINVAL;

	retval = statistic_create(&unit->stat_sizes_scsi_write, 
					unit->stat_if,
					"request_sizes_scsi_write",
					"bytes/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_sizes_scsi_read, 
					unit->stat_if,
					"request_sizes_scsi_read",
					"bytes/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_sizes_scsi_nodata, 
					unit->stat_if,
					"request_sizes_scsi_nodata",
					"bytes/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_sizes_scsi_nofit, 
					unit->stat_if,
					"request_sizes_scsi_nofit",
					"bytes/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_sizes_scsi_nomem, 
					unit->stat_if,
					"request_sizes_scsi_nomem",
					"bytes/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_sizes_timedout_write, 
					unit->stat_if,
					"request_sizes_timedout_write", 
					"bytes/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_sizes_timedout_read, 
					unit->stat_if,
					"request_sizes_timedout_read", 
					"bytes/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_sizes_timedout_nodata, 
					unit->stat_if,
					"request_sizes_timedout_nodata",
					"bytes/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_latencies_scsi_write, 
					unit->stat_if,
					"latencies_scsi_write", 
					"milliseconds/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_latencies_scsi_read, 
					unit->stat_if,
					"latencies_scsi_read", 
					"milliseconds/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_latencies_scsi_nodata, 
					unit->stat_if,
					"latencies_scsi_nodata", 
					"milliseconds/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_pending_scsi_write, 
					unit->stat_if,
					"pending_scsi_write", 
					"commands/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_pending_scsi_read, 
					unit->stat_if,
					"pending_scsi_read", 
					"commands/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_erp, 
					unit->stat_if,
					"occurrence_erp", 
					"results/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_eh_reset, 
					unit->stat_if,
					"occurrence_eh_reset", 
					"results/incidents");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_channel_latency_write,
					unit->stat_if, 
					"channel_latency_write",
					"nanosec/request");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_channel_latency_read,
					unit->stat_if, 
					"channel_latency_read",
					"nanosec/request");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_channel_latency_nodata,
					unit->stat_if, 
					"channel_latency_nodata",
					"nanosec/request");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_fabric_latency_write,
					unit->stat_if, 
					"fabric_latency_write",
					"nanosec/request");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_fabric_latency_read,
					unit->stat_if, 
					"fabric_latency_read",
					"nanosec/request");
	if (retval)
		goto err_out;

	retval = statistic_create(&unit->stat_fabric_latency_nodata,
					unit->stat_if, 
					"fabric_latency_nodata",
					"nanosec/request");
	if (retval)
		goto err_out;

	statistic_define_list(unit->stat_sizes_scsi_write, 0,
				STATISTIC_RANGE_MAX, 256);


	statistic_define_list(unit->stat_sizes_scsi_read, 0,
				STATISTIC_RANGE_MAX, 256);


	statistic_define_value(unit->stat_sizes_scsi_nodata,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);


	statistic_define_list(unit->stat_sizes_scsi_nofit, 0,
				STATISTIC_RANGE_MAX, 256);


	statistic_define_value(unit->stat_sizes_scsi_nomem, STATISTIC_RANGE_MIN,
				STATISTIC_RANGE_MAX, STATISTIC_DEF_MODE_INC);


	statistic_define_value(unit->stat_sizes_timedout_write,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);


	statistic_define_value(unit->stat_sizes_timedout_read,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);


	statistic_define_value(unit->stat_sizes_timedout_nodata,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);


	statistic_define_array(unit->stat_latencies_scsi_write, 0, 1024, 1,
				STATISTIC_DEF_SCALE_LOG2);


	statistic_define_array(unit->stat_latencies_scsi_read, 0, 1024, 1,
				STATISTIC_DEF_SCALE_LOG2);


	statistic_define_array(unit->stat_latencies_scsi_nodata, 0, 1024, 1,
				STATISTIC_DEF_SCALE_LOG2);


	statistic_define_range(unit->stat_pending_scsi_write, 0,
				STATISTIC_RANGE_MAX);


	statistic_define_range(unit->stat_pending_scsi_read, 0,
				STATISTIC_RANGE_MAX);

	statistic_define_value(unit->stat_erp,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	statistic_define_value(unit->stat_eh_reset,
				STATISTIC_RANGE_MIN, STATISTIC_RANGE_MAX,
				STATISTIC_DEF_MODE_INC);

	statistic_define_array(unit->stat_channel_latency_write,
				0, 16384000, 1000, STATISTIC_DEF_SCALE_LOG2);


	statistic_define_array(unit->stat_channel_latency_read,
				0, 16384000, 1000, STATISTIC_DEF_SCALE_LOG2);


	statistic_define_array(unit->stat_channel_latency_nodata,
				0, 16384000, 1000, STATISTIC_DEF_SCALE_LOG2);


	statistic_define_array(unit->stat_fabric_latency_write, 0,
				1024000000, 1000000, STATISTIC_DEF_SCALE_LOG2);


	statistic_define_array(unit->stat_fabric_latency_read, 0,
				1024000000, 1000000, STATISTIC_DEF_SCALE_LOG2);


	statistic_define_array(unit->stat_fabric_latency_nodata, 0,
				1024000000, 1000000, STATISTIC_DEF_SCALE_LOG2);

	return 0;
err_out:
	statistic_interface_remove(&unit->stat_if);
	return retval;
}

int zfcp_unit_statistic_unregister(struct zfcp_unit *unit)
{
	return statistic_interface_remove(&unit->stat_if);
}

#undef ZFCP_LOG_AREA
