/* Data Collection Extension to Rule-based Classification Engine (RBCE) module
 *
 * Copyright (C) Hubertus Franke, IBM Corp. 2003
 *
 * Extension to be included into RBCE to collect delay and sample information
 * Requires user daemon e.g. crbcedmn to activate.
 *
 * Latest version, more details at http://ckrm.sf.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */


/*******************************************************************************
 *
 *   User-Kernel Communication Channel (UKCC)
 *   Protocol and communication handling
 *
 ******************************************************************************/

#include <linux/moduleparam.h>
#include <linux/relayfs_fs.h>
#include <linux/ckrm_tc.h>
#include <linux/interrupt.h> /* for local_bh_*able in spin_*lock_bh */

#define PSAMPLE(pdata)    (&((pdata)->ext_data.sample))
int crbce_numbuf = 4;
module_param(crbce_numbuf, int, 0444);
MODULE_PARM_DESC(crbce_numbuf, "Number of subbuffers for UKCC operations");
int crbce_bufsize = 16;
module_param(crbce_bufsize, int, 0444);
MODULE_PARM_DESC(crbce_bufsize, "Size of subbuffer for UKCC operations (in Kbytes)");

#define CHANNEL_AUTO_CONT  0	/* this is during debugging only. It allows 
				   the module to continue sending data through 
				   the UKCC if space frees up vs. going into 
				   the recovery driven mode
				 */

enum ukcc_state {
	UKCC_OK = 0,
	UKCC_STANDBY = 1,
	UKCC_FULL = 2
};

int ukcc_channel = -1;
static enum ukcc_state chan_state = UKCC_STANDBY;
static int send_tgid = 0;
static int send_aux_event_ts = 0;

inline static int ukcc_ok(void)
{
	return (chan_state == UKCC_OK);
}

static void ukcc_cmd_deliver(int rchan_id, char *from, u32 len);
static void client_attached(void);
static void client_detached(void);

static int ukcc_fileop_notify(int rchan_id,
			      struct file *filp, enum relay_fileop fileop)
{
	static int readers = 0;
	if (fileop == RELAY_FILE_OPEN) {
		// printk("got fileop_notify RELAY_FILE_OPEN for file %p\n", 
		//		filp);
		if (readers) {
			printk("only one client allowed, backoff .... \n");
			return -EPERM;
		}
		if (!try_module_get(THIS_MODULE))
			return -EPERM;
		readers++;
		client_attached();

	} else if (fileop == RELAY_FILE_CLOSE) {
		// printk("got fileop_notify RELAY_FILE_CLOSE for file %p\n", 
		//		filp);
		client_detached();
		readers--;
		module_put(THIS_MODULE);
	}
	return 0;
}

static int create_ukcc_channel(void)
{
	static struct rchan_callbacks ukcc_callbacks = {
		.buffer_start = NULL,
		.buffer_end = NULL,
		.deliver = NULL,
		.user_deliver = ukcc_cmd_deliver,
		.needs_resize = NULL,
		.fileop_notify = ukcc_fileop_notify,
	};

	u32 channel_flags =
	    RELAY_USAGE_GLOBAL | RELAY_SCHEME_ANY | RELAY_TIMESTAMP_ANY;

	// notify on subbuffer full (through poll)
	channel_flags |= RELAY_DELIVERY_BULK;
	// channel_flags     |= RELAY_DELIVERY_PACKET;
	// avoid overwrite, otherwise recovery will be nasty...
	channel_flags |= RELAY_MODE_NO_OVERWRITE;

	ukcc_channel = relay_open(CRBCE_UKCC_NAME,
				  crbce_bufsize << 10,
				  crbce_numbuf,
				  channel_flags,
				  &ukcc_callbacks, 0, 0, 0, 0, 0, 0, NULL, 0);
	if (ukcc_channel < 0)
		printk("crbce: ukcc creation failed, errcode: %d\n",
		       ukcc_channel);
	else
		printk("crbce: ukcc created (%u KB)\n",
		       crbce_bufsize * crbce_numbuf);
	return ukcc_channel;
}

static inline void close_ukcc_channel(void)
{
	if (ukcc_channel >= 0) {
		relay_close(ukcc_channel);
		ukcc_channel = -1;
		chan_state = UKCC_STANDBY;
	}
}

#define rec_set_hdr(r,t,p)      ((r)->hdr.type = (t), (r)->hdr.pid = (p))
#define rec_set_timehdr(r,t,p,c)  (rec_set_hdr(r,t,p), \
(r)->hdr.timestamp = ckrm_jiffies_to_msecs(jiffies), (r)->hdr.cls=(unsigned long)(c) )

#if CHANNEL_AUTO_CONT

/* we only provide this for debugging.. it allows us to send records
 * based on availability in the channel when the UKCC stalles rather
 * going through the UKCC recovery protocol
 */

#define rec_send_len(r,l)						\
	do {								\
		int chan_wasok = (chan_state == UKCC_OK);		\
                int chan_isok = (relay_write(ukcc_channel,		\
					     (r),(l),-1,NULL) > 0);	\
		chan_state = chan_isok ? UKCC_OK : UKCC_STANDBY;	\
		if (chan_wasok && !chan_isok) {				\
			printk("Channel stalled\n");			\
		} else if (!chan_wasok && chan_isok) {			\
			printk("Channel continues\n");			\
		}							\
	} while (0)
#else

/* Default UKCC channel protocol. 
 * Though a UKCC buffer overflow should not happen ever, it is possible iff 
 * the user daemon stops reading for some reason. Hence we provide a simple 
 * protocol based on 3 states 
 *     UKCC_OK      := 	channel is active and properly working. When a channel 
 *		       	write fails we move to state CHAN_FULL.
 *     UKCC_FULL    := 	channel is active, but the last send_rec has failed. As
 *                      a result we will try to send an indication to the daemon
 *                     	that this has happened. When that succeeds, we move to 
 *			state UKCC_STANDBY.
 *     UKCC_STANDBY := 	we are waiting to be restarted by the user daemon 
 *
 */

static void ukcc_full(void)
{
	static spinlock_t ukcc_state_lock = SPIN_LOCK_UNLOCKED;
	/* protect transition from OK -> FULL to ensure only one record is sent,
	   rest we do not need to protect, protocol implies that. we keep the 
	   channel OK until
	*/
	int send = 0;
	spin_lock(&ukcc_state_lock);
	if ((send = (chan_state != UKCC_STANDBY)))
		chan_state = UKCC_STANDBY;	/* assume we can send */
	spin_unlock(&ukcc_state_lock);

	if (send) {
		struct crbce_ukcc_full rec;
		rec_set_timehdr(&rec, CRBCE_REC_UKCC_FULL, 0, 0);
		if (relay_write(ukcc_channel, &rec, 
				sizeof(rec), -1, NULL) <= 0) {
			/* channel remains full .. try with next one */
			spin_lock(&ukcc_state_lock);
			chan_state = UKCC_FULL;
			spin_unlock(&ukcc_state_lock);
		}
	}
}

#define rec_send_len(r,l)					       	\
	do {								\
 	        switch (chan_state) {					\
		case UKCC_OK:						\
			if (relay_write(ukcc_channel,(r),		\
				(l),-1,NULL) > 0)			\
				break;					\
		case UKCC_FULL:						\
			ukcc_full();					\
			break;						\
		default:						\
			break;						\
		}							\
	} while (0)
#endif

#define rec_send(r)	rec_send_len(r,sizeof(*(r)))


#define rec_send_ts(a,b)	rec_send_ts_val(a,b,NULL)

static inline void rec_send_ts_val(void *rec, int len, 
					struct crbce_aux_timeval *auxt)
{
	struct crbce_hdr *r = (struct crbce_hdr *)rec;
	struct crbce_aux_timeval *hdr  = (struct crbce_aux_timeval *)(rec+len);

	if (send_aux_event_ts)
		switch(r->type) {
			case CKRM_EVENT_MANUAL:
			case CKRM_EVENT_RECLASSIFY:
			case CKRM_EVENT_EXIT:
			case CKRM_EVENT_FORK:
			case CKRM_EVENT_EXEC:
			case CRBCE_REC_EXIT:
			case CRBCE_REC_TASKINFO:
				hdr->ahdr.aux_type = CRBCE_AUX_TIMEVAL;
				if (auxt)
					hdr->atv = auxt->atv;
				else
					do_gettimeofday(&hdr->atv);
				hdr->ahdr.aux_len =
					sizeof(struct crbce_aux_timeval);
				len += hdr->ahdr.aux_len;
		break;

		default:
			break;
		}

	rec_send_len(r, len);
}
			
/******************************************************************************
 *
 *  Callbacks for the CKRM engine. 
 *    In each we do the necessary classification and event record generation
 *    We generate 3 kind of records in the callback 
 *    (a) FORK              	send the pid, the class and the ppid
 *    (b) RECLASSIFICATION  	send the pid, the class and < sample data + 
 * 				delay data >
 *    (b) EXIT              	send the pid
 *
 ******************************************************************************/

int delta_mode = 0;

static inline void copy_delay(int delta, struct task_delay_info *delay, 
						struct task_struct *tsk)
{
	*delay = tsk->delays;

#if defined CONFIG_PPC_SPLPAR
	delay->splpar_total = get_tsk_accumtb(tsk);
	delay->runcpu_total = delay->splpar_total;
	if (delta)
		delay->runcpu_total -= tsk->delays.splpar_total;
	tsk->delays.splpar_total = delay->splpar_total;
#endif
}

static inline void zero_delay(struct task_delay_info *delay)
{
	memset(delay, 0, offsetof(struct task_delay_info, splpar_total));	
	/* we need to think about doing this 64-bit atomic */
}

static inline void zero_sample(struct task_sample_info *sample)
{
	memset(sample, 0, sizeof(struct task_sample_info));	
	/* we need to think about doing this 64-bit atomic */
}

static inline int check_zero(void *ptr, int len)
{
	int iszero = 1;
	int i;
	unsigned long *uptr = (unsigned long *)ptr;

	for (i = len / sizeof(unsigned long); i-- && iszero; uptr++)	
		// assume its rounded 
		iszero &= (*uptr == 0);
	return iszero;
}

static inline int check_not_zero(void *ptr, int len)
{
	int i;
	unsigned long *uptr = (unsigned long *)ptr;

	for (i = len / sizeof(unsigned long); i--; uptr++)	
		// assume its rounded 
		if (*uptr)
			return 1;
	return 0;
}

static inline int sample_changed(struct task_sample_info *s)
{
	return check_not_zero(s, sizeof(*s));
}
static inline int delay_changed(struct task_delay_info *d)
{
	return check_not_zero(d, sizeof(*d));
}

#define DONT_FORCE_SEND 0
#define FORCE_SEND 1

/*
 * caller must hold ckrm task lock
 *
 * Returns number of records filled out.
 */
static inline int
__make_task_record(struct task_struct *tsk, int event,
  		   struct ckrm_core_class *core, int send_forced,
		   struct crbce_rec_task_data *rec)
{
	struct rbce_private_data *pdata;

	pdata = RBCE_DATA(tsk);
	if (pdata == NULL)
		return 0;

	if ((send_forced == FORCE_SEND)
	    || (delta_mode == 0)
	    || sample_changed(PSAMPLE(pdata))
	    || delay_changed(&tsk->delays)) {
		rec_set_timehdr(rec, event, tsk->pid, core ? core :
				(struct ckrm_core_class *)tsk->taskclass);
		rec->sample = *PSAMPLE(pdata);
		copy_delay(delta_mode,&rec->delay, tsk);
		if (delta_mode || send_forced) {
			// on reclassify or delta mode reset the counters  
			zero_sample(PSAMPLE(pdata));
			zero_delay(&tsk->delays);
		}
		return 1;
	}
	return 0;
}

/* Caller must NOT hold tsk->ckrm_celock */
static int
send_task_record(struct task_struct *tsk, int event,
  		   struct ckrm_core_class *core, int send_forced)
{
        struct {   
                struct crbce_rec_task_data rec;
                struct crbce_aux_timeval atv;
        } rec;
	int send = 0;

	spin_lock_bh(&tsk->ckrm_celock);
	send = __make_task_record(tsk, event, core, send_forced, &rec.rec);
	spin_unlock_bh(&tsk->ckrm_celock);
	if (send == 1)
		rec_send_ts((void *)&rec, sizeof(struct crbce_rec_task_data));
	return send;
}

static inline void send_exit_notification(struct task_struct *tsk)
{
	if (ukcc_ok())
		send_task_record(tsk, CRBCE_REC_EXIT, NULL, FORCE_SEND);
}

static inline void
rbce_tc_ext_notify(int event, void *core, struct task_struct *tsk)
{
	struct {
		struct crbce_rec_fork_tgid rec;
		struct crbce_aux_timeval atv;
	} rec;

	switch (event) {
	case CKRM_EVENT_FORK:
		if (ukcc_ok()) {
			rec.rec.ppid = tsk->parent->pid;
			rec.rec.tgid = tsk->tgid;
			rec_set_timehdr(&rec.rec, CKRM_EVENT_FORK, 
							tsk->pid, core);
			if (send_tgid)
				rec_send_ts((void *)&rec,
					sizeof(struct crbce_rec_fork_tgid));
			else
				rec_send_ts((void *)&rec,
						sizeof(struct crbce_rec_fork));
		}
		break;

	case CKRM_EVENT_MANUAL:
		rbce_tc_manual(tsk);

	default:
		send_task_record(tsk, event,
				   (struct ckrm_core_class *)core,
				   FORCE_SEND);
		break;
	}
}

/*====================== end classification engine =======================*/

static void sample_task_data(unsigned long unused);

struct timer_list sample_timer = {.expires = 0,.function = sample_task_data };
unsigned long timer_interval_length = (250 * HZ) / 1000;

inline void stop_sample_timer(void)
{
	if (sample_timer.expires > 0) {
		del_timer_sync(&sample_timer);
		sample_timer.expires = 0;
	}
}

inline void start_sample_timer(void)
{
	if (timer_interval_length > 0) {
		sample_timer.expires =
		    jiffies + (timer_interval_length * HZ) / 1000;
		add_timer(&sample_timer);
	}
}

static void send_all_task_data(void)
{
	struct crbce_rec_data_delim limrec;
	struct task_struct *proc, *thread;
	int sendcnt = 0;
	int taskcnt = 0;
	limrec.is_stop = 0;
	rec_set_timehdr(&limrec, CRBCE_REC_DATA_DELIMITER, 0, 0);
	rec_send(&limrec);

	if (!ukcc_ok())
		goto skip_tasklist;

	read_lock(&tasklist_lock);
	do_each_thread(proc, thread) {
		if (ukcc_ok()) {
			taskcnt++;
			sendcnt += send_task_record(thread, CRBCE_REC_SAMPLE,
						    NULL, DONT_FORCE_SEND);
		} else
			goto done; /* No more space in buffer */
	} while_each_thread(proc, thread);
done:
	read_unlock(&tasklist_lock);
skip_tasklist:
	limrec.is_stop = 1;
	rec_set_timehdr(&limrec, CRBCE_REC_DATA_DELIMITER, 0, 0);
	rec_send(&limrec);

	// printk("send_all_task_data mode=%d t#=%d s#=%d\n",
	// 		delta_mode,taskcnt,sendcnt);
}

static void notify_class_action(struct rbce_class *cls, int action)
{
	struct crbce_class_info cinfo;
	int len;

	rec_set_timehdr(&cinfo, CRBCE_REC_CLASS_INFO, 0, cls->classobj);
	cinfo.action = action;
	len = strnlen(cls->obj.name, CRBCE_MAX_CLASS_NAME_LEN - 1);
	memcpy(&cinfo.name, cls->obj.name, len);
	cinfo.name[len] = '\0';
	len++;
	cinfo.namelen = len;

	len += sizeof(cinfo) - CRBCE_MAX_CLASS_NAME_LEN;
	rec_send_len(&cinfo, len);
}

static void send_classlist(void)
{
	struct rbce_class *cls;

	read_lock(&global_rwlock);
	list_for_each_entry(cls, &class_list, obj.link) {
		notify_class_action(cls, RBCE_NOTIFY_CLASS_ADD);
	}
	read_unlock(&global_rwlock);
}

/*
 *  resend_task_info 
 * 
 *  This function resends all essential task information to the client.
 *  Caller must NOT hold ckrm task lock.
 *
 */
static void resend_task_info(void)
{
	struct crbce_rec_data_delim limrec;
	struct task_struct *proc, *thread;
        struct {
                struct crbce_rec_fork_tgid rec;
                struct crbce_aux_timeval ahdr;
        } rec;

	send_classlist();	// first send available class information

	limrec.is_stop = 2;
	rec_set_timehdr(&limrec, CRBCE_REC_DATA_DELIMITER, 0, 0);
	rec_send(&limrec);

	if (!ukcc_ok())
		goto skip_tasklist;
	read_lock(&tasklist_lock);
	do_each_thread(proc, thread) {
		if (ukcc_ok()) {
			struct ckrm_task_class *class;
			rec.rec.ppid = thread->parent->pid;
			rec.rec.tgid = thread->tgid;
			spin_lock_bh(&thread->ckrm_celock);
			class = thread->taskclass;
			spin_unlock_bh(&thread->ckrm_celock);
			rec_set_timehdr(&rec.rec, CRBCE_REC_TASKINFO,
					thread->pid, class);
                        if (send_tgid)
                                rec_send_ts_val(&rec,
                                        sizeof(struct crbce_rec_fork_tgid),
                                        &rec.ahdr);
                        else
                                rec_send_ts_val(&rec,
                                                sizeof(struct crbce_rec_fork),
                                                &rec.ahdr);
		} else
			goto done;
	}
	while_each_thread(proc, thread);
done:
	read_unlock(&tasklist_lock);
skip_tasklist:
	limrec.is_stop = 3;
	rec_set_timehdr(&limrec, CRBCE_REC_DATA_DELIMITER, 0, 0);
	rec_send(&limrec);
}

extern int task_running_sys(struct task_struct *);

static void reset_all_private_data(void)
{
	struct task_struct *proc, *thread;

	read_lock(&tasklist_lock);
 	read_lock(&global_rwlock);
	do_each_thread(proc, thread) {
		spin_lock_bh(&thread->ckrm_celock);
 		if (RBCE_DATA(thread) == NULL)
			RBCE_DATAP(thread) = create_private_data(NULL);
 		else {
			zero_sample(PSAMPLE(RBCE_DATA(thread)));
			zero_delay(&thread->delays);
 		}
		spin_unlock_bh(&thread->ckrm_celock);
	}
	while_each_thread(proc, thread);
 	read_unlock(&global_rwlock);
	read_unlock(&tasklist_lock);
}

static void sample_task_data(unsigned long unused)
{
	struct task_struct *proc, *thread;

	int run = 0;
	int wait = 0;
	read_lock(&tasklist_lock);
	do_each_thread(proc, thread) {
		struct rbce_private_data *pdata;
		spin_lock(&thread->ckrm_celock);
		pdata = RBCE_DATA(thread);
		if (pdata == NULL){
			spin_unlock(&thread->ckrm_celock);
			continue;
		}
		if (thread->state == TASK_RUNNING) {
			if (task_running_sys(thread)) {
				PSAMPLE(pdata)->cpu_running++;
				run++;
			} else {
				PSAMPLE(pdata)->cpu_waiting++;
				wait++;
			}
		}
		/* update IO state */
		if (thread->flags & PF_IOWAIT) {
			if (thread->flags & PF_MEMIO)
				PSAMPLE(pdata)->memio_delayed++;
			else
				PSAMPLE(pdata)->io_delayed++;
		}
		spin_unlock(&thread->ckrm_celock);
	}
	while_each_thread(proc, thread);
	read_unlock(&tasklist_lock);
//      printk("sample_timer: run=%d wait=%d\n",run,wait);
	start_sample_timer();
}

inline static  void fork_event_wo_tgid(void)
{
	send_tgid = 0;
}

inline static void fork_event_with_tgid(void)
{
	send_tgid = 1;
}

inline static  void send_event_with_ts(void)
{
	send_aux_event_ts = 1;
}

inline static void send_event_wo_ts(void)
{
	send_aux_event_ts = 0;
}

static void ukcc_cmd_deliver(int rchan_id, char *from, u32 len)
{
	struct crbce_command *cmdrec = (struct crbce_command *)from;
	struct crbce_cmd_done cmdret;
	int rc = 0;

//      printk("ukcc_cmd_deliver: %d %d len=%d:%d\n",cmdrec->type, 
//		cmdrec->cmd,cmdrec->len,len);

	cmdrec->len = len;	// add this to reflection so the user doesn't 
				// accidently write the wrong length and the 
				// protocol is getting screwed up 

	if (cmdrec->type != CRBCE_REC_KERNEL_CMD) {
		rc = EINVAL;
		goto out;
	}

	switch (cmdrec->cmd) {
	case CRBCE_CMD_SET_TIMER:
		{
			struct crbce_cmd_settimer *cptr =
			    (struct crbce_cmd_settimer *)cmdrec;
			if (len != sizeof(*cptr)) {
				rc = EINVAL;
				break;
			}
			stop_sample_timer();
			timer_interval_length = cptr->interval;
			if ((timer_interval_length > 0)
			    && (timer_interval_length < 10))
				timer_interval_length = 10;   
				// anything finer can create problems 
			printk(KERN_INFO "CRBCE set sample collect timer %lu\n",
			       timer_interval_length);
			start_sample_timer();
			break;
		}
	case CRBCE_CMD_SEND_DATA:
		{
			struct crbce_cmd_send_data *cptr =
			    (struct crbce_cmd_send_data *)cmdrec;
			if (len != sizeof(*cptr)) {
				rc = EINVAL;
				break;
			}
			delta_mode = cptr->delta_mode;
			send_all_task_data();
			break;
		}
	case CRBCE_CMD_START:
		reset_all_private_data();
		chan_state = UKCC_OK;
		resend_task_info();
		break;

	case CRBCE_CMD_STOP:
		chan_state = UKCC_STANDBY;
		break;

	case CRBCE_CMD_SEND_TGID:
		fork_event_with_tgid();
		break;

	case CRBCE_CMD_NO_TGID:
		fork_event_wo_tgid();
		break;

	case CRBCE_CMD_AUXDATA_EVENT_TS:
		send_event_with_ts();
		break;

	case CRBCE_CMD_NO_AUXDATA_EVENT_TS:
		send_event_wo_ts();
		break;
	default:
		rc = EINVAL;
		break;
	}

      out:
	cmdret.hdr.type = CRBCE_REC_KERNEL_CMD_DONE;
	cmdret.hdr.cmd = cmdrec->cmd;
	cmdret.rc = rc;
	rec_send(&cmdret);
//      printk("ukcc_cmd_deliver ACK: %d %d rc=%d %d\n",cmdret.hdr.type,
//			cmdret.hdr.cmd,rc,sizeof(cmdret));
}

static void client_attached(void)
{
	printk("client [%d]<%s> attached to UKCC\n", current->pid,
	       current->comm);
	relay_reset(ukcc_channel);
}

static void client_detached(void)
{
	printk("client [%d]<%s> detached to UKCC\n", current->pid,
	       current->comm);
	chan_state = UKCC_STANDBY;
	stop_sample_timer();
	fork_event_wo_tgid();
	send_event_wo_ts();
	relay_reset(ukcc_channel);
}

static int init_rbce_ext_pre(void)
{
	int rc;

	rc = create_ukcc_channel();
	return ((rc < 0) ? rc : 0);
}

static int init_rbce_ext_post(void)
{
	init_timer(&sample_timer);
	return 0;
}

static void exit_rbce_ext(void)
{
	stop_sample_timer();
	close_ukcc_channel();
}
